"""Create simple packets with variable lengths"""

import struct

import numpy as np

out_name = "var_length_packets_double_varfield_with_footer.bin"
include_footer = True
num_packets = 10
AP_ID = 0x08E2
PACKET_ID = int("0001100000000000", 2) + AP_ID

packet = b""

data1_length = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]
data2_length = [5, 7, 2, 8, 3, 41, 42, 1, 3, 4]

assert len(data1_length) == num_packets
assert len(data2_length) == num_packets


for packet_num in range(num_packets):
    # Determine packet length
    packet_length = 0
    packet_length += 1
    packet_length += 2 * data1_length[packet_num]
    packet_length += 1
    packet_length += 2 * data2_length[packet_num]
    packet_length += -1

    if include_footer:
        packet_length += 2

    # Add packet header
    this_packet = struct.pack(">HHH", PACKET_ID, packet_num, packet_length)

    #  data1 array contents start with the array size value and increment by one (e.g. [2, 3], [3, 4, 5]) # noqa: E501
    this_packet += struct.pack("B", data1_length[packet_num])
    for j in range(data1_length[packet_num]):
        this_packet += struct.pack(">H", data1_length[packet_num] + j)

    #  data2 array contents start with the zero and increment to the data2_length
    this_packet += struct.pack("B", data2_length[packet_num])
    for j in range(data2_length[packet_num]):
        this_packet += struct.pack(">H", j)

    # Footer is two bytes at the end
    if include_footer:
        this_packet += struct.pack(">H", 1)

    print(f"packet #{packet_num}: {np.frombuffer(this_packet, dtype='uint8')}")
    packet += this_packet

if include_footer:
    expected_length = (
        8 * num_packets + 2 * np.sum(data1_length) + 2 * np.sum(data2_length) + 2 * num_packets
    )
else:
    expected_length = (
        6 * num_packets + 2 * np.sum(data1_length) + 2 * np.sum(data2_length) + 2 * num_packets
    )

print(f"File length is {len(packet)}. Expected length is {expected_length}.")

f = open(out_name, "wb")
f.write(packet)
f.close()



"""Tests for complex byte orderings, eg. 3412.

The test data covers all XTCE orderings requested in 
  https://github.com/CCSDSPy/ccsdspy/discussions/110

See also:
  ccsdspy/tests/data/byte_order/byte_order_packets.py
"""

import glob
import itertools
import os
import string

import numpy as np
import pytest

from ccsdspy import (
    PacketField,
    PacketArray,
    VariableLength,
    FixedLength,
)

# Current directory of the module we are in.
CUR_DIR = os.path.dirname(os.path.abspath(__file__))

# The test_byte_order function takes combinational variations of arguments.
# These constants and the itertools.product() function create these
# combinations.
PARAM_BYTE_ORDER_FILES = glob.glob(f"{CUR_DIR}/data/byte_order/byteorder*.bin")
PARAM_CLASSES = [FixedLength, VariableLength]
PARAM_DECODEMETHODS = ["array", "fields", "multiarray"]
PARAM_OPTIONS = list(itertools.product(PARAM_BYTE_ORDER_FILES, PARAM_CLASSES, PARAM_DECODEMETHODS))


@pytest.mark.parametrize("bin_file,cls,decode_method", PARAM_OPTIONS)
def test_byte_order(bin_file, cls, decode_method):
    """Test loading a .bin file and matching its contents to its .csv
    counterpart.

    See also:
      ccsdspy/tests/data/byte_order/byte_order_packets.py
    """
    csv_file = bin_file.replace(".bin", ".csv")
    assert os.path.exists(bin_file)
    assert os.path.exists(csv_file)

    byte_order = os.path.basename(bin_file).split(".")[0].split("_")[1]
    assert all((char in string.digits) for char in byte_order)

    # Parse CSV file
    expected_file_data = []

    with open(csv_file) as fh:
        for line in fh:
            line_values = [int(v) for v in line.strip().split(",")]
            expected_file_data.append(line_values)

    expected_file_data = np.array(expected_file_data)

    if decode_method in ("array", "multiarray", "expand"):
        shape = {
            "array": 80,
            "multiarray": (40, 2),
            "expand": "expand",
        }[decode_method]

        # Build Packet with PacketArray instance
        packet = cls(
            [
                PacketArray(
                    name="body",
                    data_type="uint",
                    array_shape=shape,
                    bit_length=8 * len(byte_order),
                    byte_order=byte_order,
                )
            ]
        )
        result = packet.load(bin_file)

        for got_data, expected_data in zip(result["body"], expected_file_data):
            if decode_method == "multiarray":
                expected_data = np.array(expected_data).reshape(shape)

            np.testing.assert_array_equal(got_data, expected_data)
    else:
        # Build Packet with individual fields
        packet_fields = []

        for i in range(80):
            packet_fields.append(
                PacketField(
                    name=f"body{i}",
                    data_type="uint",
                    bit_length=8 * len(byte_order),
                    byte_order=byte_order,
                )
            )
        packet = cls(packet_fields)
        result = packet.load(bin_file)

        for pkt_num, expected_data in enumerate(expected_file_data):
            got_data = [result[f"body{i}"][pkt_num] for i in range(80)]
            np.testing.assert_array_equal(got_data, expected_data)


def test_expanding():
    """Simple test for expanding fields."""
    bin_file = f"{CUR_DIR}/data/byte_order/byteorder_1234.bin"
    test_byte_order(bin_file, VariableLength, "expand")



"""Create simple packets with variable lengths"""

import struct

import numpy as np

out_name = "var_length_packets_with_footer.bin"
include_footer = True
num_packets = 10
AP_ID = 0x08E2
PACKET_ID = int("0001100000000000", 2) + AP_ID

packet = b""

data_length = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]

for packet_num in range(num_packets):
    packet_length = 2 * data_length[packet_num] - 1
    if include_footer:
        packet_length += 2

    this_packet = struct.pack(">HHH", PACKET_ID, packet_num, packet_length)
    #  data array contents start with the array size value and increment by one (e.g. [2, 3], [3, 4, 5]) # noqa: E501
    for j in range(data_length[packet_num]):
        this_packet += struct.pack(">H", data_length[packet_num] + j)

    if include_footer:
        this_packet += struct.pack(">H", 1)

    print(f"packet #{packet_num}: {np.frombuffer(this_packet, dtype='uint8')}")
    packet += this_packet

if include_footer:
    expected_length = 8 * num_packets + 2 * np.sum(data_length)
else:
    expected_length = 6 * num_packets + 2 * np.sum(data_length)

print(f"Packet length is {len(packet)}. Expected length is {expected_length}.")

f = open(out_name, "wb")
f.write(packet)
f.close()



"""Run end-to-end test of decoding fixed-length packets
using the test data in the data/hs directory.
"""

import csv
import glob
import os
import numpy as np

import pytest

from .. import FixedLength, VariableLength, PacketField, PacketArray
from ..constants import BITS_PER_BYTE


def _run_apid_test(apid, cls):
    """Driver for running an APID test. Each apidXXX directory under
    `data/hs` contains:

            defs.csv     -- packet definition with conversions.
            xxxx.tlm     -- binary CCSDS file
    """
    # Setup paths for the definitions, truth, and CCSDS files
    # in the APID directory.
    dir_path = os.path.dirname(os.path.realpath(__file__))
    apid_dir = os.path.join(dir_path, "data", "hs", f"apid{apid:03d}")

    defs_file_path = os.path.join(apid_dir, "defs.csv")
    ccsds_file_path = glob.glob(os.path.join(apid_dir, "*.tlm")).pop()

    assert all(
        os.path.exists(path)
        for path in (
            apid_dir,
            defs_file_path,
            ccsds_file_path,
        )
    )

    # Load the definitions, and test that they parse.  We have not prepared
    # a truth reference for this set of test data.
    defs = _load_apid_defs(defs_file_path)
    _decode_ccsds_file(ccsds_file_path, defs, cls)


def _load_apid_defs(defs_file_path):
    """Load APID definitions (defs.csv) and return a Table"""
    table_dict = {
        "name": [],
        "data_type": [],
        "bit_offset": [],
        "bit_length": [],
    }

    # Loop through CSV lines, read all as strings.
    with open(defs_file_path) as fh:
        reader = csv.reader(fh)
        first_line = True

        for row in reader:
            if first_line:
                first_line = False
                continue

            for key, row_value in zip(table_dict.keys(), row):
                table_dict[key].append(row_value)

    # Change from string to final type.
    table_dict["name"] = [name.upper() for name in table_dict["name"]]
    table_dict["bit_offset"] = [int(n) for n in table_dict["bit_offset"]]
    table_dict["bit_length"] = [int(n) for n in table_dict["bit_length"]]

    return table_dict


def _decode_ccsds_file(ccsds_file_path, defs, cls):
    pkt_fields = []

    for key, data_type, bit_length in zip(defs["name"], defs["data_type"], defs["bit_length"]):
        pkt_fields.append(PacketField(name=key, data_type=data_type, bit_length=bit_length))

    # Skip first two, which specify the primary header
    pkt = cls(pkt_fields[2:])
    decoded = pkt.load(ccsds_file_path)

    return decoded


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_hs_apid001(cls):
    _run_apid_test(1, cls)


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_hs_apid010(cls):
    _run_apid_test(10, cls)


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_hs_apid035(cls):
    _run_apid_test(35, cls)


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_hs_apid130(cls):
    _run_apid_test(130, cls)


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_hs_apid251(cls):
    _run_apid_test(251, cls)


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_hs_apid895(cls):
    _run_apid_test(895, cls)


def test_hs_apid035_PacketArray():
    # Make FixedLength for normal packet
    dir_path = os.path.dirname(os.path.realpath(__file__))
    apid_dir = os.path.join(dir_path, "data", "hs", "apid035")
    defs_file_path = os.path.join(apid_dir, "defs.csv")
    dict_normal_defs = _load_apid_defs(defs_file_path)

    pkt_fields = []
    tmp = (
        dict_normal_defs["name"],
        dict_normal_defs["data_type"],
        dict_normal_defs["bit_length"],
    )

    for key, data_type, bit_length in zip(*tmp):
        pkt_fields.append(PacketField(name=key, data_type=data_type, bit_length=bit_length))

    normal_pkt = FixedLength(pkt_fields)

    # Make FixedLength for array packet
    fill_length = sum(f._bit_length for f in pkt_fields if "[" not in f._name)

    array_pkt = FixedLength(
        [
            PacketField(name="unused", data_type="fill", bit_length=fill_length),
            PacketArray(name="PKT35_FLT_SIN_2H", data_type="float", bit_length=32, array_shape=8),
        ]
    )

    # Compare data
    ccsds_file_path = glob.glob(os.path.join(apid_dir, "*.tlm")).pop()
    normal_result = normal_pkt.load(ccsds_file_path)
    array_result = array_pkt.load(ccsds_file_path)

    for i in range(8):
        assert np.all(
            normal_result[f"PKT35_FLT_SIN_2H[{i}]"] == array_result["PKT35_FLT_SIN_2H"][:, i]
        )
        # These sin values should be between [0, 1]
        assert np.abs(normal_result[f"PKT35_FLT_SIN_2H[{i}]"].min() - -1.0) < 1e-3
        assert np.abs(normal_result[f"PKT35_FLT_SIN_2H[{i}]"].max() - +1.0) < 1e-3



"""Utils for the CCSDSPy package."""

__author__ = "Daniel da Silva <mail@danieldasilva.org>"

from io import BytesIO
import warnings

import numpy as np

from . import VariableLength, PacketArray
from .constants import BITS_PER_BYTE, PRIMARY_HEADER_NUM_BYTES
from . import decode


def get_packet_total_bytes(primary_header_bytes):
    """[autogenerated]"""
    # Wrap around internal function from decode module
    return decode._get_packet_total_bytes(primary_header_bytes)


def get_packet_apid(primary_header_bytes):
    """[autogenerated]"""
    # Wrap around internal function from decode module
    return decode._get_packet_apid(primary_header_bytes)


# Copy __doc__ from decode module for functions which wrap decode module functions
# so that the doc need not be repeated in two places.
get_packet_total_bytes.__doc__ = decode._get_packet_total_bytes.__doc__
get_packet_apid.__doc__ = decode._get_packet_apid.__doc__


def iter_packet_bytes(file, include_primary_header=True):
    """Iterate through packets as raw bytes objects, in the order they appear in a file.

    This function works with mixed files containing multiple APIDs, which may
    include both fixed length and variable length packets.

    If end of last packet doesn't align with end of file, a warning is issued.

    Parameters
    ----------
    file : str, file-like
      Path to file on the local file system, or file-like object
    include_primary_header : bool
      If set to False, excludes the primary header bytes (the first six)

    Yields
    ------
    packet_bytes : bytes
       Bytes associated with each packet as it appears in the file. When
       include_primary_header=False, the primary header bytes are excluded.
    """
    if hasattr(file, "read"):
        file_bytes = np.frombuffer(file.read(), "u1")
    else:
        file_bytes = np.fromfile(file, "u1")

    offset = 0

    if include_primary_header:
        delta_idx = 0
    else:
        delta_idx = PRIMARY_HEADER_NUM_BYTES

    while offset < len(file_bytes):
        packet_nbytes = get_packet_total_bytes(
            file_bytes[offset : offset + PRIMARY_HEADER_NUM_BYTES].tobytes()
        )
        packet_bytes = file_bytes[offset + delta_idx : offset + packet_nbytes].tobytes()

        yield packet_bytes

        offset += packet_nbytes

    if offset != len(file_bytes):
        missing_bytes = offset - len(file_bytes)
        message = (
            f"File appears truncated-- missing {missing_bytes} byte (or " "maybe garbage at end)"
        )
        warnings.warn(message)


def split_packet_bytes(file, include_primary_header=True):
    """Retrieve a list of bytes objects corresponding to each packet in a file.

    This function works with mixed files containing multiple APIDs, which may
    include both fixed length and variable length packets.

    If end of last packet doesn't align with end of file, a warning is issued.

    Parameters
    ----------
    file : str, file-like
      Path to file on the local file system, or file-like object
    include_primary_header : bool
      If set to False, excludes the primary header bytes (the first six)

    Returns
    -------
    packet_bytes : list of bytes
       List of bytes objects associated with each packet as it appears in the
       file. When include_primary_header=False, each byte object will have its
       primary header bytes excluded.
    """
    return list(iter_packet_bytes(file, include_primary_header=include_primary_header))


def read_primary_headers(file):
    """Read primary header fields and return contents as a dictionary
    of arrays.

    This function works with mixed files containing multiple APIDs, which may
    include both fixed length and variable length packets.

    Parameters
    ----------
    file : str, file-like
      Path to file on the local file system, or file-like object

    Returns
    -------
    header_arrays : dict, string to NumPy array
       Dictionary mapping header names to NumPy arrays. The header names are:
       `CCSDS_VERSION_NUMBER`, `CCSDS_PACKET_TYPE`, `CCSDS_SECONDARY_FLAG`,
       `CCSDS_SEQUENCE_FLAG`, `CCSDS_APID`, `CCSDS_SEQUENCE_COUNT`,
       `CCSDS_PACKET_LENGTH`
    """
    pkt = VariableLength(
        [
            PacketArray(
                name="unused", data_type="uint", bit_length=BITS_PER_BYTE, array_shape="expand"
            )
        ]
    )

    header_arrays = pkt.load(file, include_primary_header=True)
    del header_arrays["unused"]

    return header_arrays


def split_by_apid(mixed_file, valid_apids=None):
    """Split a stream of mixed APIDs into separate streams by APID.

    This works with a mix of both fixed length and variable length packets.

    Parameters
    ----------
    mixed_file: str, file-like
       Path to file on the local file system, or file-like object
    valid_apids: list of int, None
       Optional list of valid APIDs. If specified, warning will be issued when
       an APID is encountered outside this list.

    Returns
    -------
    stream_by_apid : dict, int to :py:class:`~io.BytesIO`
      Dictionary mapping integer apid number to BytesIO instance with the file
      pointer at the beginning of the stream.
    """
    # If not None, convert valid_apids to set for faster lookup times
    if valid_apids is not None:
        valid_apids = set(valid_apids)

    stream_by_apid = {}

    for packet_bytes in iter_packet_bytes(mixed_file):
        apid = get_packet_apid(packet_bytes[:PRIMARY_HEADER_NUM_BYTES])

        if valid_apids is not None and apid not in valid_apids:
            warnings.warn(f"Found unknown APID {apid}")

        if apid not in stream_by_apid:
            stream_by_apid[apid] = BytesIO()

        stream_by_apid[apid].write(packet_bytes)

    for stream in stream_by_apid.values():
        stream.seek(0)

    return stream_by_apid


def count_packets(file, return_missing_bytes=False):
    """Count the number of packets in a file and check if there are any
    missing bytes in the last packet.

    This function works with mixed files containing multiple APIDs, which may
    include both fixed length and variable length packets. When used with
    multiple APIDs, it simply returns the total number of packets of any APID.

    If end of last packet doesn't align with end of file, a warning is issued.

    Parameters
    ----------
    file : str, file-like
      Path to file on the local file system, or file-like object
    return_missing_bytes : bool, optional
      Also return the number of missing bytes at the end of the file. This
      is the number of bytes which would need to be added to the file to
      complete the last packet expected (as set by the packet length in
      the last packet's primary header).

    Returns
    -------
    num_packets : int
       Number of complete packets in the file
    missing_bytes : int, optional
      The number of bytes which would need to be added to the file to
      complete the last packet expected (as set by the packet length in
      the last packet's primary header).
    """
    if hasattr(file, "read"):
        file_bytes = np.frombuffer(file.read(), "u1")
    else:
        file_bytes = np.fromfile(file, "u1")

    offset = 0
    num_packets = 0

    while offset < len(file_bytes):
        packet_nbytes = get_packet_total_bytes(
            file_bytes[offset : offset + PRIMARY_HEADER_NUM_BYTES].tobytes()
        )
        offset += packet_nbytes
        num_packets += 1

    missing_bytes = offset - len(file_bytes)

    if offset != len(file_bytes):
        missing_bytes = offset - len(file_bytes)
        message = (
            f"File appears truncated-- missing {missing_bytes} byte (or " "maybe garbage at end)"
        )
        warnings.warn(message)

    if return_missing_bytes:
        return num_packets, missing_bytes
    else:
        return num_packets



import numpy as np
import struct

np.random.seed(0)

# Obtained from github discusison #110
TEST_BYTE_ORDERS = {
    1: ["1"],
    2: ["12", "21"],
    3: ["123", "321"],
    4: ["1234", "2143", "3412", "4321"],
    6: ["123456"],
    7: ["1234567"],
    8: ["12345678", "78563412", "87654321"],
}

NUM_PACKETS = 10
PACKET_ARRAY_LENGTH = 80


def main():
    for _, byte_order_list in TEST_BYTE_ORDERS.items():
        for byte_order in byte_order_list:
            print(f"Byte Order: {byte_order}")
            write_packet(byte_order)


def write_packet(byte_order):
    bin_fname = f"byteorder_{byte_order}.bin"
    csv_fname = f"byteorder_{byte_order}.csv"
    apid = 0x08E2
    packet_id = int("0001100000000000", 2) + apid
    file_bytes = b""
    csv_lines = []

    for i in range(NUM_PACKETS):
        packet_length = len(byte_order) * PACKET_ARRAY_LENGTH - 1
        this_packet = struct.pack(">HHH", packet_id, i, packet_length)

        array = np.random.randint(
            low=0, high=2 ** (8 * len(byte_order)) - 1, size=PACKET_ARRAY_LENGTH, dtype=np.uint64
        )

        array_bytes = array.copy()
        array_bytes.dtype = np.uint8
        array_bytes = array_bytes.reshape((array.size, 8))

        for j in range(PACKET_ARRAY_LENGTH):
            cur_bytes = [array_bytes[j, int(byte_num) - 1] for byte_num in byte_order]
            cur_bytes.reverse()
            for b in cur_bytes:
                this_packet += struct.pack("B", b)

        file_bytes += this_packet
        csv_lines.append(",".join(map(str, array.tolist())) + "\n")

    expected_length = 6 * NUM_PACKETS + len(byte_order) * NUM_PACKETS * PACKET_ARRAY_LENGTH
    print(f"  Packet length is {len(file_bytes)}. Expected length is {expected_length}.")

    with open(bin_fname, "wb") as bin_fh:
        bin_fh.write(file_bytes)

    print(f"  Wrote {bin_fname}")

    with open(csv_fname, "w") as csv_fh:
        csv_fh.writelines(csv_lines)

    print(f"  Wrote {csv_fname}")


if __name__ == "__main__":
    main()



"""This class hold the implementation of the converter system, which applies 
post-process to decoded packet fields. This post-processing includes applying
linear/polynomial calibration curves, dictionary replacement, and time parsing.
"""

from datetime import datetime, timedelta

import numpy as np

__all__ = [
    "EnumConverterMissingKey",
    "Converter",
    "PolyConverter",
    "LinearConverter",
    "EnumConverter",
    "DatetimeConverter",
    "StringifyBytesConverter",
]


class EnumConverterMissingKey(RuntimeError):
    """During conversion a value was encountered which did not have a
    corresponding key in the replacement dictionary.
    """


class Converter:
    """Base class for all converter objects.

    This class is extended to create converters, and users may extend this
    class to write their own custom converters.

    To write a converter, one must create a subclass and override either the method
    `convert(*field_arrays)`. This method implements the conversion for an entire
    sequence of decoded packet field values in a single call.
    """

    def __init__(self):
        raise NotImplementedError("This is a base class not meant to be instantiated directly")

    def convert(self, field_array):
        """Convert a sequence of decoded packet field values.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least one dimension

        Returns
        -------
        converted_field_array : NumPy array
            converted form of the decoded packet field values
        """
        raise NotImplementedError("This method must be overridden by a subclass")


class PolyConverter(Converter):
    """Post-processing conversion which applies calibration using a series
    of coefficients ordered from highest power to intercept.
    """

    def __init__(self, coeffs):
        """Instantiate a PolyConverter object

        Parameters
        ----------
        coeffs : list of float
           Polynomial coefficients ordered from highest power to intercept.
        """
        self._coeffs = coeffs

    def convert(self, field_array):
        """Apply the polynomial conversion.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least one dimension

        Returns
        -------
        converted : NumPy array
            converted form of the decoded packet field values
        """
        converted = np.zeros(field_array.shape, dtype=np.float64)

        for power, coeff in enumerate(reversed(self._coeffs)):
            converted += coeff * field_array**power

        return converted


class LinearConverter(PolyConverter):
    """Post-processing conversion which applies a linear (y=mx+b)
    transformation.
    """

    def __init__(self, slope, intercept):
        """Instantiate a LinearConverter"""
        super().__init__([slope, intercept])


class EnumConverter(Converter):
    """Post-processing conversion for applying dictionary replacement of
    integers to strings.

    If during conversion a value is encountered which does not have a
    corresponding key in the replacement dictionary, then a
    `:py:class:`~ccsdspy.converters.EnumConverterMissingKey` exception
    will be thrown.
    """

    def __init__(self, replace_dict):
        """Initialize a EnumConverter.

        Parameters
        ----------
        replace_dict : dict of int to string
           Replacement dictionary mapping integer values to string values

        Raises
        ------
        TypeError
           Either one of the keys of the replacement dictionary is not an
           integer, or one of the values is not a string.
        """
        self._replace_dict = replace_dict

        for key, value in replace_dict.items():
            if not isinstance(key, int):
                raise TypeError(
                    f"Found key in EnumConverter replacement dictionary that is "
                    f"not an integer: {repr(key)}"
                )
            if not isinstance(value, str):
                raise TypeError(
                    f"Found value in EnumConverter replacement dictionary that is "
                    f"not a string: {repr(value)}"
                )

    def convert(self, field_array):
        """Apply the enum replacement conversion.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least one dimension

        Returns
        -------
        converted : NumPy array
            converted form of the decoded packet field values
        """
        converted = np.zeros(field_array.shape, dtype=object)
        converted_mask = np.zeros(field_array.shape, dtype=bool)

        for key, value in self._replace_dict.items():
            converted[field_array == key] = value
            converted_mask[field_array == key] = True

        if not converted_mask.all():
            missing_keys = field_array[~converted_mask].tolist()

            raise EnumConverterMissingKey(
                f"The following were encountered which did not have "
                f"corresponding keys in the replacement dictionary: "
                f"{repr(missing_keys)}"
            )

        return converted


class DatetimeConverter(Converter):
    """Post-processing conversion for converting timestamp fields to datetime
    instances, computed using offset(s) from a reference time.

    This class supports the offsets stored in multiple input fields, for example
    where one field is a coarse time (e.g. seconds) and a second field is a fine
    time (e.g. nanoseconds). To use multiple input fields, pass a tuple of input
    field names when this converter is added to the packet.
    """

    _VALID_UNITS = (
        "days",
        "hours",
        "minutes",
        "seconds",
        "milliseconds",
        "microseconds",
        "nanoseconds",
    )
    _MILLISECONDS_PER_SECOND = 1_000
    _MICROSECONDS_PER_SECOND = 1_000_000
    _NANOSECONDS_PER_SECOND = 1_000_000_000

    def __init__(self, since, units):
        """Initialize a DatetimeConverter

        Parameters
        ----------
        since : datetime
          Reference datetime. The time stored in the field(s) is considered an
          offset to this reference. If this has timezone information attached to
          it, so will the converted datetimes.
        units : str or tuple of str
          Units string of tuples of units strings for the offset of each
          input field.  Valid units are "days", "hours", "minutes",
          "seconds", "milliseconds", "microseconds", and "nanoseconds".

        Raises
        ------
        TypeError
          One of the input arguments is not of the correct type
        ValueError
          One or more of the units are invalid
        """
        if not isinstance(since, datetime):
            raise TypeError("Argument 'since' must be an instance of datetime")

        if isinstance(units, str):
            units_tuple = (units,)
        elif isinstance(units, tuple):
            units_tuple = units
        else:
            raise TypeError("Argument 'units' must be either a string or tuple")

        if not (set(units_tuple) <= set(self._VALID_UNITS)):
            raise ValueError("One or more units are invalid")

        self._since = since
        self._units = units_tuple

    def convert(self, *field_arrays):
        """Apply the datetime conversion.

        Parameters
        ----------
        field_arrays : list of NumPy array
          list of decoded packet field values, each must have at least one
          dimension

        Returns
        -------
        converted : NumPy array of object (holding datetimes)
          converted form of the decoded packet field values

        Raises
        ------
        ValueError
          Too many or too few units were provided, as compared to the
          input field arrays sent.
        """
        assert len(field_arrays) > 0, "Must have at least one input field"

        converted = []

        for field_values in zip(*field_arrays):
            converted_time = self._since

            for unit, offset_raw in zip(self._units, field_values):
                offset_raw = float(offset_raw)

                if unit == "days":
                    converted_time += timedelta(days=offset_raw)
                elif unit == "hours":
                    converted_time += timedelta(hours=offset_raw)
                elif unit == "minutes":
                    converted_time += timedelta(minutes=offset_raw)
                elif unit == "seconds":
                    converted_time += timedelta(seconds=offset_raw)
                elif unit == "milliseconds":
                    converted_time += timedelta(seconds=offset_raw / self._MILLISECONDS_PER_SECOND)
                elif unit == "microseconds":
                    converted_time += timedelta(seconds=offset_raw / self._MICROSECONDS_PER_SECOND)
                elif unit == "nanoseconds":
                    converted_time += timedelta(seconds=offset_raw / self._NANOSECONDS_PER_SECOND)

            converted.append(converted_time)

        converted = np.array(converted, dtype=object)

        return converted


class StringifyBytesConverter(Converter):
    """Post-processing conversion which converts byte arrays or multi-byte
    numbers to strings in numeric representations such as binary, hexadecimal,
    or octal.

    To convert individual bytes, the input field should be defined as a
    `~ccsdspy.PacketArray` constructed with `data_type="uint"` and
    `bit_length=8`. Otherwise, each element is converted as a single entity.

    If the field is an array, the shape of the array is retained. The strings
    generated are not padded to a fixed length.

    The converted strings contain prefixes such as `0b` (binary), `0x` (hex),
    or `0o` (octal). If the number is signed and negative, the prefixes change
    to `-0b` (binary), `-0x` (hex), or `-0o` (octal).
    """

    def __init__(self, format="hex"):
        """Instantiate a StringifyBytesConverter object

        Parameters
        ----------
        format : {"bin", "hex", "oct"}
           Format used to encode the bytes in a string.
        """
        if format not in ("bin", "hex", "oct"):
            raise ValueError(
                "The format= keyword passed to StringifyBytesConverter "
                f"must be either 'bin', 'hex', or 'oct'. Got {repr(format)}"
            )

        self._format = format

    def _stringify_number(self, num, nbytes):
        """Internal helper method to convert a number to a string.

        Parameters
        ----------
        number : int
           A single number to convert to string

        Returns
        --------
        as_string : the byte converted to a string using the format
           specified when this object was created.
        """
        if self._format == "bin":
            return bin(num)
        elif self._format == "hex":
            return hex(num)
        else:
            return oct(num)

    def convert(self, field_array):
        """Apply the conversion.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least two dimensions

        Returns
        -------
        converted : NumPy array
            converted form of the converted packet field values
        """
        # field_arrays may either be a 1-D array, or an N-D array where N>1
        # (this includes jagged arrays where the outer array is of
        # dtype=object). These are implemented separately.
        ndims = len(field_array.shape)

        if ndims == 1 and field_array.dtype != object:
            converted = []

            for num in field_array:
                as_string = self._stringify_number(num, field_array.itemsize)
                converted.append(as_string)
        else:
            converted = []

            for i in range(field_array.shape[0]):
                cur_array_flat = field_array[i].flatten()
                n_items = cur_array_flat.shape[0]
                cur_shape = field_array[i].shape

                # Loop over elements, converting individually
                curr_array_strings = []

                for element in cur_array_flat:
                    as_string = self._stringify_number(element, cur_array_flat.itemsize)
                    curr_array_strings.append(as_string)

                # Put back into original array shape
                curr_array_strings = np.array(curr_array_strings, dtype=object).reshape(cur_shape)
                converted.append(curr_array_strings)

        converted = np.array(converted, dtype=object)

        return converted



"""Tests for the ccsdspy.packet_types module"""

__author__ = "Daniel da Silva"

import io
import os.path
import struct

import numpy as np
import pytest

from .. import FixedLength, VariableLength, PacketField, PacketArray
from ..constants import BITS_PER_BYTE
from ..packet_types import _get_fields_csv_file, _parse_csv_array_shape

dir_path = os.path.dirname(os.path.realpath(__file__))
packet_def_dir = os.path.join(dir_path, "data", "packet_def")
csv_file_4col = os.path.join(packet_def_dir, "extended_csv_4col.csv")
csv_file_3col = os.path.join(packet_def_dir, "basic_csv_3col.csv")
csv_file_4col_with_array = os.path.join(packet_def_dir, "extended_csv_4col_with_array.csv")
csv_file_3col_with_array = os.path.join(packet_def_dir, "basic_csv_3col_with_array.csv")
csv_file_3col_with_expand = os.path.join(packet_def_dir, "basic_csv_3col_with_expand.csv")
csv_file_3col_with_reference = os.path.join(packet_def_dir, "basic_csv_3col_with_reference.csv")
csv_file_3col_with_all = os.path.join(packet_def_dir, "basic_csv_3col_with_all.csv")

hs_packet_dir = os.path.join(dir_path, "data", "hs")
random_binary_file = os.path.join(
    hs_packet_dir, "apid001", "SSAT1_2015-180-00-00-00_2015-180-01-59-58_1_1_sim.tlm"
)
random_packet_def = os.path.join(hs_packet_dir, "apid001", "defs.csv")


def test_FixedLength_initializer_copies_field_list():
    """Tests that the FixedLengthPacket initializer stores a copy of the
    provided fields list.
    """
    fields = [PacketField(name="mnemonic", data_type="uint", bit_length=BITS_PER_BYTE)]
    pkt = FixedLength(fields)
    assert pkt._fields is not fields


def test_get_fields_csv_file_3col():
    """Test that the csv reader parses the 3 column test file correctly."""

    names = []
    data_types = []
    bit_lengths = []
    with open(csv_file_3col, "r") as fp:
        lines = fp.readlines()
        num_lines = len(lines)
        for line in lines[1:]:  # skip the header row
            cols = line.split(",")
            names.append(cols[0].strip())
            data_types.append(cols[1].strip())
            bit_lengths.append(int(cols[2]))

    fields = _get_fields_csv_file(csv_file_3col)

    assert num_lines - 1 == len(fields)
    for i, this_field in enumerate(fields):
        assert this_field._name == names[i]
        assert this_field._data_type == data_types[i]
        assert this_field._bit_length == bit_lengths[i]


def test_get_fields_csv_file_4col():
    """Test that the csv reader parses the 4 column test file correctly."""

    names = []
    data_types = []
    bit_lengths = []
    bit_offsets = []
    with open(csv_file_4col, "r") as fp:
        lines = fp.readlines()
        num_lines = len(lines)
        for line in lines[1:]:  # skip the header row
            cols = line.split(",")
            names.append(cols[0].strip())
            data_types.append(cols[1].strip())
            bit_lengths.append(int(cols[2]))
            bit_offsets.append(int(cols[3]))

    fields = _get_fields_csv_file(csv_file_4col)

    assert num_lines - 1 == len(fields)
    for i, this_field in enumerate(fields):
        assert this_field._name == names[i]
        assert this_field._data_type == data_types[i]
        assert this_field._bit_length == bit_lengths[i]
        assert this_field._bit_offset == bit_offsets[i]


def test_FixedLength_from_file():
    """Test that from_file returns a FixedLength instance"""
    assert isinstance(FixedLength.from_file(csv_file_3col), FixedLength)
    assert isinstance(FixedLength.from_file(csv_file_4col), FixedLength)
    assert isinstance(FixedLength.from_file(csv_file_3col_with_array), FixedLength)
    assert isinstance(FixedLength.from_file(csv_file_4col_with_array), FixedLength)


@pytest.mark.parametrize("filename", ["boo.txt", "great.zip"])
def test_FixedLength_from_file_not_supported(filename):
    """Test that if given an unsupported filetype raises an error"""
    with pytest.raises(ValueError):
        FixedLength.from_file(filename)


@pytest.mark.parametrize(
    "shape_str, expected_value",
    [
        ("uint(4)", 4),
        ("uint(1, 2)", (1, 2)),
        ("uint(expand)", "expand"),
        ("uint(OPMODE)", "OPMODE"),
    ],
)
def test_parse_csv_array_shape(shape_str, expected_value):
    assert _parse_csv_array_shape(shape_str) == expected_value


def test_parse_csv_array_shape_fails_on_invalid_shape_str():
    with pytest.raises(ValueError):
        _parse_csv_array_shape("uint(4, FIELD)")


def test_VariableLength_from_file():
    """Test that from_file returns a VariableLength instance"""
    assert isinstance(VariableLength.from_file(csv_file_3col_with_expand), VariableLength)
    assert isinstance(VariableLength.from_file(csv_file_3col_with_reference), VariableLength)
    assert isinstance(VariableLength.from_file(csv_file_3col_with_all), VariableLength)


@pytest.mark.parametrize(
    "cls,numpy_dtype,ccsdspy_data_type,ccsdspy_bit_length,array_order,include_bit_offset",
    [
        (FixedLength, ">f4", "float", 32, "C", False),
        (FixedLength, ">f4", "float", 32, "F", False),
        (FixedLength, ">u2", "uint", 16, "C", False),
        (FixedLength, ">u2", "uint", 16, "F", False),
        (FixedLength, ">u8", "uint", 64, "C", False),
        (FixedLength, ">u8", "uint", 64, "F", False),
        (FixedLength, ">i4", "int", 32, "C", False),
        (FixedLength, ">i4", "int", 32, "F", False),
        (FixedLength, ">i4", "int", 32, "F", True),
        (VariableLength, ">f4", "float", 32, "C", False),
        (VariableLength, ">f4", "float", 32, "F", False),
        (VariableLength, ">u2", "uint", 16, "C", False),
        (VariableLength, ">u2", "uint", 16, "F", False),
        (VariableLength, ">u8", "uint", 64, "C", False),
        (VariableLength, ">u8", "uint", 64, "F", False),
        (VariableLength, ">i4", "int", 32, "C", False),
        (VariableLength, ">i4", "int", 32, "F", False),
    ],
)
def test_multidimensional_array(
    cls,
    numpy_dtype,
    ccsdspy_data_type,
    ccsdspy_bit_length,
    array_order,
    include_bit_offset,
):
    """Test the PacketArray class with a multidimensional array.

    See test_hs.py for a test with a 1-dimensional array
    """
    # Each packet holds a 32x4 array where the array of the k'th packet has:
    #   arr[i, j] = k * (2 * i + j)
    num_packets = 10
    arrays = {}

    for k in range(num_packets):
        arrays[k] = np.zeros((32, 4), dtype=numpy_dtype)

        for i in range(32):
            for j in range(4):
                arrays[k][i, j] = k * (2 * i + j)

    # Generate packets
    apid = 0x08E2
    packet_id = int("0001100000000000", 2) + apid
    packet_stream = b""

    for packet_num in range(num_packets):
        packet_length = arrays[packet_num].nbytes - 1
        this_packet = struct.pack(">HHH", packet_id, packet_num, packet_length)
        this_packet += arrays[packet_num].tobytes(order=array_order)
        packet_stream += this_packet

    assert len(packet_stream) == num_packets * (6 + arrays[0].nbytes)

    # Build fixed length parse packet stream
    if include_bit_offset:
        bit_offset = 48
    else:
        bit_offset = None

    pkt = cls(
        [
            PacketArray(
                name="array",
                data_type=ccsdspy_data_type,
                bit_length=ccsdspy_bit_length,
                array_shape=(32, 4),
                array_order=array_order,
                bit_offset=bit_offset,
            )
        ]
    )
    results = pkt.load(io.BytesIO(packet_stream))

    # Check results
    for k in range(num_packets):
        for i in range(32):
            for j in range(4):
                assert results["array"][k, i, j] == k * (2 * i + j)


def test_fixed_length_rejects_expanding():
    with pytest.raises(ValueError):
        FixedLength(
            [
                PacketArray(
                    name="array",
                    array_shape="expand",
                    data_type="uint",
                    bit_length=BITS_PER_BYTE,
                )
            ]
        )


def test_variable_length_rejects_multiple_expanding():
    with pytest.raises(ValueError):
        FixedLength(
            [
                PacketArray(
                    name="array1",
                    array_shape="expand",
                    data_type="uint",
                    bit_length=BITS_PER_BYTE,
                ),
                PacketArray(
                    name="array2",
                    array_shape="expand",
                    data_type="uint",
                    bit_length=BITS_PER_BYTE,
                ),
            ]
        )


def test_variable_length_rejects_bit_offset():
    with pytest.raises(ValueError):
        FixedLength(
            [
                PacketArray(
                    name="array1",
                    array_shape="expand",
                    data_type="uint",
                    bit_length=BITS_PER_BYTE,
                    bit_offset=32,
                ),
            ]
        )


def test_load_without_moving_file_buffer_pos():
    """Tests that load(..., reset_file_obj=True) works as intended."""
    pkts = FixedLength.from_file(random_packet_def)
    with open(random_binary_file, "rb") as fp:
        pos = fp.tell()
        pkts.load(fp, reset_file_obj=True)
        assert pos == fp.tell()
